import glob

import yaml
from beartype.typing import List, Optional, Set
from jinja2 import Environment, StrictUndefined, meta
from jinja2.exceptions import UndefinedError
from loguru import logger

from dbt_jobs_as_code.schemas.config import Config


class LoadingJobsYAMLError(Exception):
    pass


def load_job_configuration(config_files: List[str], vars_file: Optional[List[str]]) -> Config:
    """Load the job configuration set in a YAML file into a Config object

    Can be a non-templated YAML or a templated one for which we need to replace Jinja values
    """
    if vars_file:
        config = _load_yaml_with_template(config_files, vars_file)
    else:
        config = _load_yaml_no_template(config_files)

    if config.get("jobs", {}) == {}:
        return Config(jobs={})

    date_config = [job.get("schedule", {}).get("date", None) for job in config["jobs"].values()]
    time_config = [job.get("schedule", {}).get("time", None) for job in config["jobs"].values()]

    if any(date_config):
        logger.warning(
            f"⚡️ There is some date config under 'schedule > date' in your YML. This data is auto generated and should be deleted. Only cron is supported in the config."
        )
    if any(time_config):
        logger.warning(
            f"⚡️ There is some time config under 'schedule > time' in your YML. This data is auto generated and should be deleted. Only cron is supported in the config."
        )

    for identifier, job in config.get("jobs", {}).items():
        job["identifier"] = identifier

    return Config(**config)


def _load_yaml_no_template(config_files: List[str]) -> dict:
    """Load a job YAML file into a Config object"""

    combined_config = {}
    for config_file in config_files:
        with open(config_file) as f:
            config_string = f.read()

            jinja_vars = _get_jinja_variables(config_string)
            if jinja_vars:
                raise LoadingJobsYAMLError(
                    f"{config_file} is a templated YAML file. Please remove the variables {jinja_vars} or provide the variables values."
                )

            config = yaml.safe_load(config_string)
            if config:
                # Merge the jobs from each file into combined_config
                if config.get("jobs", {}) != {}:
                    if "jobs" not in combined_config:
                        combined_config["jobs"] = {}
                    combined_config["jobs"].update(config["jobs"])
                # Merge any other top-level keys
                for key, value in config.items():
                    if key != "jobs":
                        combined_config[key] = value

    return combined_config


def _load_yaml_with_template(config_files: List[str], vars_file: List[str]) -> dict:
    """Load a job YAML file into a Config object"""
    # Load and merge vars files
    template_vars_values = {}
    for vars_path in vars_file:
        with open(vars_path) as f:
            # we load the vars. if there is no data in them we set it to an empty dict
            vars_data = yaml.safe_load(f) or {}
            # Check for duplicate variables
            for key in vars_data:
                if key in template_vars_values:
                    raise LoadingJobsYAMLError(
                        f"Variable '{key}' is defined multiple times in vars files"
                    )
            template_vars_values.update(vars_data)

    # Load and combine config files
    combined_config = {}
    env = Environment(undefined=StrictUndefined)

    for config_path in config_files:
        with open(config_path) as f:
            config_string_unrendered = f.read()
            template = env.from_string(config_string_unrendered)

            try:
                config_string_rendered = template.render(template_vars_values)
            except UndefinedError as e:
                raise LoadingJobsYAMLError(
                    f"Some variables didn't have a value: {e.message}."
                ) from e

            config = yaml.safe_load(config_string_rendered)
            if config:
                # Merge the jobs from each file
                if "jobs" in config:
                    if "jobs" not in combined_config:
                        combined_config["jobs"] = {}
                    combined_config["jobs"].update(config["jobs"])
                # Merge any other top-level keys
                for key, value in config.items():
                    if key != "jobs":
                        combined_config[key] = value

    return combined_config


def _get_jinja_variables(input: str) -> Set[str]:
    """Get the variables from a Jinja template"""
    env = Environment()
    parsed_input = env.parse(input)
    return meta.find_undeclared_variables(parsed_input)


def resolve_file_paths(
    config_pattern: Optional[str], vars_pattern: Optional[str] = None
) -> tuple[List[str], List[str]]:
    """
    Resolve glob patterns to lists of file paths.

    Args:
        config_pattern: Glob pattern for config files
        vars_pattern: Optional glob pattern for vars files

    Returns:
        Tuple of (config_files, vars_files)

    Raises:
        LoadingJobsYAMLError: If no files match a pattern when pattern is provided
    """
    if not config_pattern:
        return [], []

    config_files = glob.glob(config_pattern)
    if not config_files:
        raise LoadingJobsYAMLError(f"No files found matching pattern: {config_pattern}")

    vars_files = []
    if vars_pattern:
        vars_files = glob.glob(vars_pattern)
        if not vars_files:
            raise LoadingJobsYAMLError(f"No files found matching pattern: {vars_pattern}")

    return config_files, vars_files



import json

import pytest
from jsonschema import validate
from ruamel.yaml import YAML

from dbt_jobs_as_code.exporter.export import export_jobs_yml
from dbt_jobs_as_code.schemas.common_types import (
    Date,
    Execution,
    Schedule,
    Settings,
    Time,
    Triggers,
)
from dbt_jobs_as_code.schemas.job import JobDefinition


@pytest.fixture
def base_job_definition():
    return JobDefinition(
        account_id=1,
        project_id=1,
        environment_id=1,
        name="Test Job",
        settings={},
        run_generate_sources=False,
        execute_steps=[],
        generate_docs=False,
        schedule={"cron": "0 14 * * 0,1,2,3,4,5,6"},
        triggers={},
    )


def test_export_jobs_yml(capsys):
    """Test that the export_jobs_yml function works as expected."""

    expected_output = """
# yaml-language-server: $schema=https://raw.githubusercontent.com/dbt-labs/dbt-jobs-as-code/main/src/dbt_jobs_as_code/schemas/load_job_schema.json

jobs:
  import_1:
    account_id: 50
    project_id: 100
    environment_id: 248
    dbt_version:
    name: Event Model
    settings:
      threads: 4
      target_name: production
    execution:
      timeout_seconds: 0
    deferring_job_definition_id:
    deferring_environment_id:
    run_generate_sources: true
    execute_steps:
      - dbt source freshness
      - dbt build
    generate_docs: false
    schedule:
      cron: 0 14 * * 0,1,2,3,4,5,6
    triggers:
      github_webhook: false
      git_provider_webhook: false
      schedule: false
      on_merge: false
    description: ''
    run_compare_changes: false
    compare_changes_flags: --select state:modified
    job_type: scheduled
    triggers_on_draft_pr: false
    job_completion_trigger_condition:
    custom_environment_variables: []
""".strip()

    job_def = JobDefinition(
        id=31,
        identifier=None,
        account_id=50,
        project_id=100,
        environment_id=248,
        dbt_version=None,
        name="Event Model",
        settings=Settings(threads=4, target_name="production"),
        execution=Execution(timeout_seconds=0),
        deferring_job_definition_id=None,
        deferring_environment_id=None,
        run_generate_sources=True,
        execute_steps=[
            "dbt source freshness",
            "dbt build",
        ],
        generate_docs=False,
        schedule=Schedule(
            cron="0 14 * * 0,1,2,3,4,5,6",
            date=Date(type="custom_cron", cron="0 14 * * 0,1,2,3,4,5,6"),
            time=Time(type="every_hour", interval=1),
        ),
        triggers=Triggers(
            github_webhook=False,
            git_provider_webhook=False,
            schedule=False,
        ),
        state=1,
        custom_environment_variables=[],
        description="",
    )

    export_jobs_yml([job_def])
    captured = capsys.readouterr()
    assert captured.out.strip() == expected_output

    # checking that the output YAML is a valid job
    with open("src/dbt_jobs_as_code/schemas/load_job_schema.json") as f:
        schema = f.read()

    yaml = YAML(typ="safe")
    yaml_data = yaml.load(captured.out.strip())
    validate(instance=yaml_data, schema=json.loads(schema))


def test_export_jobs_yml_with_identifier(base_job_definition, capsys):
    # Create a job with identifier
    job_with_identifier = base_job_definition.model_copy()
    job_with_identifier.identifier = "existing_identifier"

    # Create a job without identifier
    job_without_identifier = base_job_definition.model_copy()

    jobs = [job_with_identifier, job_without_identifier]

    # Export jobs and capture output
    export_jobs_yml(jobs)
    captured = capsys.readouterr()

    # Parse the YAML output (skipping the first two lines which contain the schema)
    yaml = YAML()
    exported_jobs = yaml.load("\n".join(captured.out.split("\n")[2:]))

    # Verify the job keys
    assert "existing_identifier" in exported_jobs["jobs"]
    assert "import_2" in exported_jobs["jobs"]

    # Verify the job contents
    assert exported_jobs["jobs"]["existing_identifier"]["name"] == "Test Job"
    assert exported_jobs["jobs"]["import_2"]["name"] == "Test Job"


def test_export_jobs_yml_with_linked_id(base_job_definition, capsys):
    # Create a job with both identifier and id
    job = base_job_definition.model_copy()
    job.identifier = "test_identifier"
    job.id = 123

    # Export with include_linked_id=True
    export_jobs_yml([job], include_linked_id=True)
    captured = capsys.readouterr()

    yaml = YAML()
    exported_jobs = yaml.load("\n".join(captured.out.split("\n")[2:]))

    # Verify linked_id is included and matches the id
    assert exported_jobs["jobs"]["test_identifier"]["linked_id"] == 123

    # Export with include_linked_id=False
    export_jobs_yml([job], include_linked_id=False)
    captured = capsys.readouterr()

    yaml = YAML()
    exported_jobs = yaml.load("\n".join(captured.out.split("\n")[2:]))

    # Verify linked_id is not included
    assert "linked_id" not in exported_jobs["jobs"]["test_identifier"]



from unittest.mock import Mock

import pytest

from dbt_jobs_as_code.importer import fetch_jobs, get_account_id
from dbt_jobs_as_code.schemas.job import JobDefinition


def test_get_account_id():
    # Test account ID from direct input
    assert get_account_id(None, 123) == 123

    # Test missing both inputs
    with pytest.raises(ValueError):
        get_account_id(None, None)


def test_fetch_jobs():
    mock_dbt = Mock()

    # Mock job objects
    mock_job1 = JobDefinition(
        id=1,
        name="Job 1",
        project_id=100,
        environment_id=200,
        account_id=300,
        settings={},
        run_generate_sources=False,
        execute_steps=[],
        generate_docs=False,
        schedule={"cron": "0 14 * * 0,1,2,3,4,5,6"},
        triggers={},
    )
    mock_job2 = JobDefinition(
        id=2,
        name="Job 2",
        project_id=100,
        environment_id=200,
        account_id=300,
        settings={},
        run_generate_sources=False,
        execute_steps=[],
        generate_docs=False,
        schedule={"cron": "0 14 * * 0,1,2,3,4,5,6"},
        triggers={},
    )

    # Set return values for mocks
    mock_dbt.get_job.side_effect = [mock_job1, mock_job2]
    mock_dbt.get_jobs.return_value = [mock_job1, mock_job2]

    # Test fetch with only job IDs
    jobs = fetch_jobs(mock_dbt, [1, 2], [], [])
    assert mock_dbt.get_job.call_count == 2
    assert len(jobs) == 2

    # Test fetch with project IDs
    jobs = fetch_jobs(mock_dbt, [1], [100], [])
    mock_dbt.get_jobs.assert_called_with(project_ids=[100], environment_ids=[])
    assert len(jobs) == 1



import re

from beartype.typing import Any, List, Optional
from pydantic import BaseModel, ConfigDict, Field

from dbt_jobs_as_code.schemas.common_types import (
    Execution,
    JobCompletionTriggerCondition,
    Schedule,
    Settings,
    Triggers,
    field_mandatory_int_allowed_as_string_in_schema,
    field_optional_int_allowed_as_string_in_schema,
)
from dbt_jobs_as_code.schemas.custom_environment_variable import CustomEnvironmentVariable


# Main model for loader
class JobDefinition(BaseModel):
    """A definition for a dbt Cloud job."""

    linked_id: Optional[int] = Field(
        default=None,
        description="The ID of the job in dbt Cloud that we want to link. Only used for the 'link' command.",
    )
    id: Optional[int] = None
    identifier: Optional[str] = Field(
        default=None,
        description="The internal job identifier for the job for dbt-jobs-as-code. Will be added at the end of the job name.",
    )
    account_id: int = field_mandatory_int_allowed_as_string_in_schema
    project_id: int = field_mandatory_int_allowed_as_string_in_schema
    environment_id: int = field_mandatory_int_allowed_as_string_in_schema
    dbt_version: Optional[str] = None
    name: str
    settings: Settings
    execution: Execution = Execution()
    deferring_job_definition_id: Optional[int] = field_optional_int_allowed_as_string_in_schema
    deferring_environment_id: Optional[int] = field_optional_int_allowed_as_string_in_schema
    run_generate_sources: bool
    execute_steps: List[str]
    generate_docs: bool
    schedule: Schedule
    triggers: Triggers
    description: str = ""
    state: int = 1
    run_compare_changes: bool = False
    compare_changes_flags: str = "--select state:modified"
    # we don't want to enforce the list in case we add more, but still want to get those in the JSON schema
    job_type: str = Field(
        json_schema_extra={"enum": ["scheduled", "merge", "ci", "other"]},
        default="scheduled",
    )
    triggers_on_draft_pr: bool = False
    job_completion_trigger_condition: Optional[JobCompletionTriggerCondition] = None
    custom_environment_variables: List[CustomEnvironmentVariable] = Field(
        default=[],
        json_schema_extra={
            "items": {
                "type": "object",
                "patternProperties": {
                    "^DBT_": {
                        "oneOf": [
                            {"type": "string"},
                            {"type": "number"},
                            {"type": "boolean"},
                        ]
                    },
                },
                "additionalProperties": False,
            },
        },
        description="Dictionary of custom environment variables name and value for the job. The env var name must start with DBT_.",
    )

    def __init__(self, **data: Any):
        # Check if `name` includes an identifier. If yes, set the identifier in the object. Remove the identifier from
        # the name.
        matches = re.search(r"\[\[([a-zA-Z0-9_-]+)\]\]", data["name"])
        if matches is not None:
            data["identifier"] = matches.groups()[0]
            data["name"] = data["name"].replace(f" [[{data['identifier']}]]", "")

        # Rewrite custom environment variables to include account and project id
        environment_variables = data.get("custom_environment_variables", None)
        if environment_variables:
            data["custom_environment_variables"] = [
                {
                    "name": list(variable.keys())[0],
                    "value": list(variable.values())[0],
                    "project_id": data["project_id"],
                    "account_id": data["account_id"],
                }
                for variable in environment_variables
            ]
        else:
            data["custom_environment_variables"] = []

        super().__init__(**data)

    def to_payload(self):
        """Create a dbt Cloud API payload for a JobDefinition."""

        # Rewrite the job name to embed the job ID from job.yml
        payload = self.model_copy()
        # if there is an identifier, add it to the name
        # otherwise, it means that we are "unlinking" the job from the job.yml
        if self.identifier:
            payload.name = f"{self.name} [[{self.identifier}]]"
        return payload.model_dump_json(
            exclude={"linked_id", "identifier", "custom_environment_variables"}
        )

    def to_load_format(self, include_linked_id: bool = False):
        """Generate a dict following our YML format to dump as YML later."""

        self.linked_id = self.id
        exclude_dict = {
            "identifier": True,
            "schedule": {
                "date": True,
                "time": True,
            },
            "id": True,
            "custom_environment_variables": True,
            "state": True,
        }
        if not include_linked_id:
            exclude_dict["linked_id"] = True

        data = self.model_dump(exclude=exclude_dict)
        data["custom_environment_variables"] = []
        for env_var in self.custom_environment_variables:
            data["custom_environment_variables"].append({env_var.name: env_var.value})
        return data

    def to_url(self, account_url: str) -> str:
        """Generate a URL for the job in dbt Cloud."""
        return f"{account_url}/deploy/{self.account_id}/projects/{self.project_id}/jobs/{self.id}"


class JobMissingFields(JobDefinition):
    """This class can be used to identify when there are new fields added to jobs
    We don't add the not needed fields to the JobDefinition model to prevent the tool from breaking with any API change
    """

    model_config = ConfigDict(extra="forbid")

    # when adding fields we also need to update the test for pytest

    # TODO: Add to JobDefinition model when the feature is out
    integration_id: Optional[int] = None
    run_lint: Optional[bool] = None
    errors_on_lint_failure: Optional[bool] = None

    # Unneeded read-only fields
    raw_dbt_version: Optional[str] = None
    created_at: str = ""
    updated_at: str = ""
    deactivated: bool
    run_failure_count: int = False
    lifecycle_webhooks: bool = False
    lifecycle_webhooks_url: Optional[str] = None
    is_deferrable: bool = False
    generate_sources: bool = False
    cron_humanized: str = ""
    next_run: Optional[str] = ""
    next_run_humanized: Optional[str] = ""
    is_system: bool
    account: Any
    project: Any
    environment: Any
    most_recent_run: Any
    most_recent_completed_run: Any



import glob
import os
import string
from collections import Counter

from beartype import BeartypeConf, BeartypeStrategy, beartype
from beartype.typing import Callable, List
from loguru import logger
from pydantic import BaseModel
from rich.table import Table

from dbt_jobs_as_code.client import DBTCloud, DBTCloudException
from dbt_jobs_as_code.loader.load import LoadingJobsYAMLError, load_job_configuration
from dbt_jobs_as_code.schemas import check_env_var_same, check_job_mapping_same
from dbt_jobs_as_code.schemas.job import JobDefinition

# Dynamically create a new @nobeartype decorator disabling type-checking.
nobeartype = beartype(conf=BeartypeConf(strategy=BeartypeStrategy.O0))


class Change(BaseModel):
    """Describes what a given change is and how to apply it."""

    identifier: str
    type: str
    action: str
    proj_id: int
    env_id: int
    sync_function: Callable
    parameters: dict

    def __str__(self):
        return f"{self.action.upper()} {string.capwords(self.type)} {self.identifier}"

    def apply(self):
        self.sync_function(**self.parameters)


class ChangeSet(BaseModel):
    """Store the set of changes to be displayed or applied."""

    root: List[Change] = []
    apply_success: bool = True

    def __iter__(self):
        return iter(self.root)

    def append(self, change: Change):
        self.root.append(change)

    def __str__(self):
        list_str = [str(change) for change in self.root]
        return "\n".join(list_str)

    def to_table(self) -> Table:
        """Return a table representation of the changeset."""

        table = Table(title="Changes detected")

        table.add_column("Action", style="cyan", no_wrap=True)
        table.add_column("Type", style="magenta")
        table.add_column("ID", style="green")
        table.add_column("Proj ID", style="yellow")
        table.add_column("Env ID", style="red")

        for change in self.root:
            table.add_row(
                change.action.upper(),
                string.capwords(change.type),
                change.identifier,
                str(change.proj_id),
                str(change.env_id),
            )

        return table

    def __len__(self):
        return len(self.root)

    def apply(self):
        for change in self.root:
            try:
                change.apply()
            except DBTCloudException:
                self.apply_success = False


# Don't bear type this function as we do some odd things in tests
@nobeartype
def filter_config(
    defined_jobs: dict[str, JobDefinition], project_ids: List[int], environment_ids: List[int]
) -> dict[str, JobDefinition]:
    """Filters the config based on the inputs provided for project ids and environment ids."""
    removed_job_ids: set[str] = set()
    if len(environment_ids) != 0:
        for job_id, job in defined_jobs.items():
            if job.environment_id not in environment_ids:
                removed_job_ids.add(job_id)
                logger.warning(
                    f"For Job# {job.identifier}, environment_id(s) provided as arguments does not match the ID's in Jobs YAML file!!!"
                )
    if len(project_ids) != 0:
        for job_id, job in defined_jobs.items():
            if job.project_id not in project_ids:
                removed_job_ids.add(job_id)
                logger.warning(
                    f"For Job# {job.identifier}, project_id(s) provided as arguments does not match the ID's in Jobs YAML file!!!"
                )
    return {job_id: job for job_id, job in defined_jobs.items() if job_id not in removed_job_ids}


def _check_no_duplicate_job_identifier(remote_jobs: List[JobDefinition]):
    """Check if there are duplicate job identifiers in dbt Cloud.

    If so, raise some error logs"""
    count_identifiers = Counter(
        [job.identifier for job in remote_jobs if job.identifier is not None]
    )
    multiple_counts = {item: count for item, count in count_identifiers.items() if count > 1}

    jobs_affected = [job for job in remote_jobs if job.identifier in multiple_counts]
    for job in jobs_affected:
        logger.error(
            f"The job {job.id} has a duplicate identifier '{job.identifier}' in dbt Cloud: {job.to_url(account_url=os.environ.get('DBT_BASE_URL', 'https://cloud.getdbt.com'))}"
        )


def _check_single_account_id(defined_jobs: List[JobDefinition]):
    """Check if there are duplicate job identifiers in dbt Cloud.

    If so, raise some error logs"""
    count_account_ids = Counter([job.account_id for job in defined_jobs])

    if len(count_account_ids) > 1:
        logger.error(
            f"There are different account_id in the jobs YAML file: {count_account_ids.keys()}"
        )


def build_change_set(
    config: str,
    yml_vars: str,
    disable_ssl_verification: bool,
    project_ids: List[int],
    environment_ids: List[int],
    limit_projects_envs_to_yml: bool = False,
):
    """Compares the config of YML files versus dbt Cloud.
    Depending on the value of no_update, it will either update the dbt Cloud config or not.

    CONFIG is the path to your jobs.yml config file.
    """
    # Get list of files matching the glob pattern
    config_files = glob.glob(config)
    if not config_files:
        logger.error(f"No files found matching pattern: {config}")
        return ChangeSet()

    yml_vars_files = glob.glob(yml_vars) if yml_vars else None

    try:
        configuration = load_job_configuration(config_files, yml_vars_files)
    except (LoadingJobsYAMLError, KeyError) as e:
        logger.error(f"Error loading jobs YAML file ({type(e).__name__}): {e}")
        exit(1)

    if limit_projects_envs_to_yml:
        # if limit_projects_envs_to_yml is True, we keep all the YML jobs
        defined_jobs = configuration.jobs
        # and only the remote jobs with project_id and environment_id existing in the job YML file are considered
        project_ids = list({job.project_id for job in defined_jobs.values()})
        environment_ids = list({job.environment_id for job in defined_jobs.values()})

    else:
        # If a project_id or environment_id is passed in as a parameter (one or multiple), check if these match the ID's in Jobs YAML file, otherwise add a warning and continue the process
        unfiltered_defined_jobs = configuration.jobs
        defined_jobs = filter_config(unfiltered_defined_jobs, project_ids, environment_ids)

    if len(defined_jobs) == 0:
        logger.warning(
            "No jobs found in the Jobs YAML file after filtering based on the project_id and environment_id provided as arguments!!!"
        )
        return ChangeSet()

    _check_single_account_id(list(defined_jobs.values()))

    dbt_cloud = DBTCloud(
        account_id=list(defined_jobs.values())[0].account_id,
        api_key=os.environ.get("DBT_API_KEY"),
        base_url=os.environ.get("DBT_BASE_URL", "https://cloud.getdbt.com"),
        disable_ssl_verification=disable_ssl_verification,
    )

    cloud_jobs = dbt_cloud.get_jobs(project_ids=project_ids, environment_ids=environment_ids)
    _check_no_duplicate_job_identifier(cloud_jobs)
    tracked_jobs = {job.identifier: job for job in cloud_jobs if job.identifier is not None}

    dbt_cloud_change_set = ChangeSet()

    # Use sets to find jobs for different operations
    shared_jobs = set(defined_jobs.keys()).intersection(set(tracked_jobs.keys()))
    created_jobs = set(defined_jobs.keys()) - set(tracked_jobs.keys())
    deleted_jobs = set(tracked_jobs.keys()) - set(defined_jobs.keys())

    # Update changed jobs
    logger.info("Detected {count} existing jobs.", count=len(shared_jobs))
    for identifier in shared_jobs:
        logger.info("Checking for differences in {identifier}", identifier=identifier)
        if not check_job_mapping_same(
            source_job=defined_jobs[identifier], dest_job=tracked_jobs[identifier]
        ):
            dbt_cloud_change = Change(
                identifier=identifier,
                type="job",
                action="update",
                proj_id=defined_jobs[identifier].project_id,
                env_id=defined_jobs[identifier].environment_id,
                sync_function=dbt_cloud.update_job,
                parameters={"job": defined_jobs[identifier]},
            )
            dbt_cloud_change_set.append(dbt_cloud_change)
            defined_jobs[identifier].id = tracked_jobs[identifier].id

    # Create new jobs
    logger.info("Detected {count} new jobs.", count=len(created_jobs))
    for identifier in created_jobs:
        dbt_cloud_change = Change(
            identifier=identifier,
            type="job",
            action="create",
            proj_id=defined_jobs[identifier].project_id,
            env_id=defined_jobs[identifier].environment_id,
            sync_function=dbt_cloud.create_job,
            parameters={"job": defined_jobs[identifier]},
        )
        dbt_cloud_change_set.append(dbt_cloud_change)

    # Remove Deleted Jobs
    logger.info("Detected {count} deleted jobs.", count=len(deleted_jobs))
    for identifier in deleted_jobs:
        dbt_cloud_change = Change(
            identifier=identifier,
            type="job",
            action="delete",
            proj_id=tracked_jobs[identifier].project_id,
            env_id=tracked_jobs[identifier].environment_id,
            sync_function=dbt_cloud.delete_job,
            parameters={"job": tracked_jobs[identifier]},
        )
        dbt_cloud_change_set.append(dbt_cloud_change)

    # -- ENV VARS --
    # Now that we have replicated all jobs we can get their IDs for further API calls
    mapping_job_identifier_job_id = dbt_cloud.build_mapping_job_identifier_job_id(cloud_jobs)
    logger.debug(f"Mapping of job identifier to id: {mapping_job_identifier_job_id}")

    # Replicate the env vars from the YML to dbt Cloud
    for job in defined_jobs.values():
        if job.identifier in mapping_job_identifier_job_id:  # the job already exists
            job_id = mapping_job_identifier_job_id[job.identifier]
            all_env_vars_for_job = dbt_cloud.get_env_vars(project_id=job.project_id, job_id=job_id)
            for env_var_yml in job.custom_environment_variables:
                env_var_yml.job_definition_id = job_id
                same_env_var, env_var_id = check_env_var_same(
                    source_env_var=env_var_yml, dest_env_vars=all_env_vars_for_job
                )
                if not same_env_var:
                    dbt_cloud_change = Change(
                        identifier=f"{job.identifier}:{env_var_yml.name}",
                        type="env var overwrite",
                        action="update",
                        proj_id=job.project_id,
                        env_id=job.environment_id,
                        sync_function=dbt_cloud.update_env_var,
                        parameters={
                            "project_id": job.project_id,
                            "job_id": job_id,
                            "custom_env_var": env_var_yml,
                            "env_var_id": env_var_id,
                        },
                    )
                    dbt_cloud_change_set.append(dbt_cloud_change)

        else:  # the job doesn't exist yet so it doesn't have an ID
            for env_var_yml in job.custom_environment_variables:
                dbt_cloud_change = Change(
                    identifier=f"{job.identifier}:{env_var_yml.name}",
                    type="env var overwrite",
                    action="create",
                    proj_id=job.project_id,
                    env_id=job.environment_id,
                    sync_function=dbt_cloud.update_env_var,
                    parameters={
                        "project_id": job.project_id,
                        "job_id": None,
                        "custom_env_var": env_var_yml,
                        "env_var_id": None,
                        "yml_job_identifier": job.identifier,
                    },
                )
                dbt_cloud_change_set.append(dbt_cloud_change)

    # Delete the env vars from dbt Cloud that are not in the yml
    for job in defined_jobs.values():
        # we only delete env var overwrite if the job already exists
        if job.identifier in mapping_job_identifier_job_id:
            job_id = mapping_job_identifier_job_id[job.identifier]

            # We get the env vars from dbt Cloud, now that the YML ones have been replicated
            env_var_dbt_cloud = dbt_cloud.get_env_vars(project_id=job.project_id, job_id=job_id)

            # And we get the list of env vars defined for a given job in the YML
            env_vars_for_job = [env_var.name for env_var in job.custom_environment_variables]

            for env_var, env_var_val in env_var_dbt_cloud.items():
                # If the env var is not in the YML but is defined at the "job" level in dbt Cloud, we delete it
                if env_var not in env_vars_for_job and env_var_val.id:
                    logger.info(f"{env_var} not in the YML file but in the dbt Cloud job")
                    dbt_cloud_change = Change(
                        identifier=f"{job.identifier}:{env_var}",
                        type="env var overwrite",
                        action="delete",
                        proj_id=job.project_id,
                        env_id=job.environment_id,
                        sync_function=dbt_cloud.delete_env_var,
                        parameters={
                            "project_id": job.project_id,
                            "env_var_id": env_var_val.id,
                        },
                    )
                    dbt_cloud_change_set.append(dbt_cloud_change)

    # Filtering out the change set, if project_id(s), environment_id(s) are passed as arguments to function
    # TODO: Confirm if this is the desired functionality, remove otherwise
    logger.debug(f"dbt cloud change set: {dbt_cloud_change_set}")

    return dbt_cloud_change_set



from dataclasses import dataclass
from typing import Optional

from dbt_jobs_as_code.client import DBTCloud, DBTCloudException
from dbt_jobs_as_code.schemas.job import JobDefinition


@dataclass
class LinkableCheck:
    can_be_linked: bool
    message: str
    linked_job: Optional[JobDefinition] = None


def can_be_linked(
    job_identifier: str, job_definition: JobDefinition, dbt_cloud: DBTCloud
) -> LinkableCheck:
    if job_definition.linked_id is None:
        return LinkableCheck(
            False, f"Job '{job_identifier}' doesn't have an ID in YAML. It cannot be linked"
        )

    try:
        cloud_job = dbt_cloud.get_job(job_id=job_definition.linked_id)
    except DBTCloudException as e:
        return LinkableCheck(
            False,
            f"Job {job_definition.linked_id} doesn't exist in dbt Cloud. It cannot be linked",
        )

    if cloud_job.identifier is not None:
        return LinkableCheck(
            False,
            f"Job {job_definition.linked_id} is already linked with the identifier {cloud_job.identifier}. You should unlink it before if you want to link it to a new identifier.",
        )

    return LinkableCheck(True, "", cloud_job)



import os

import pytest
from dbt_jobs_as_code.client import DBTCloud
from pydantic_core import ValidationError

DBT_ACCOUNT_ID = int(os.getenv("DBT_ACCOUNT_ID", 0))
DBT_API_KEY = os.getenv("DBT_API_KEY")
DBT_BASE_URL = os.getenv("DBT_BASE_URL", "")
DBT_JOB_ID = int(os.getenv("DBT_JOB_ID", 0))


def test_new_job_fields():
    client = DBTCloud(
        account_id=DBT_ACCOUNT_ID,
        api_key=DBT_API_KEY,
        base_url=DBT_BASE_URL,
    )

    try:
        parsed_job = client.get_job_missing_fields(
            job_id=DBT_JOB_ID,
        )
    except ValidationError as e:
        pytest.fail(
            f"ValidationError was raised: {e.errors()}\n\nThis means that a field needs to be explicitly excluded in JobMissingFields or added to the JobDefinition model."
        )
    except Exception as e:
        pytest.fail(f"A generic error was raised: {e}")

    if not parsed_job:
        pytest.fail("No job was parsed, check if it still exists")



from beartype.typing import Any, Dict, List, Literal, Optional
from croniter import croniter
from pydantic import BaseModel, Field, field_serializer, field_validator


def set_one_of_string_integer(schema: Dict[str, Any]):
    schema.pop("type", None)
    schema["oneOf"] = [{"type": "string"}, {"type": "integer"}]


def set_any_of_string_integer_null(schema: Dict[str, Any]):
    schema.pop("type", None)
    schema["anyOf"] = [{"type": "string"}, {"type": "integer"}, {"type": "null"}]


field_mandatory_int_allowed_as_string_in_schema = Field(
    json_schema_extra=set_one_of_string_integer
)
field_optional_int_allowed_as_string_in_schema = Field(
    default=None, json_schema_extra=set_any_of_string_integer_null
)


class Execution(BaseModel):
    timeout_seconds: int = 0


def set_any_of_string_boolean(schema: Dict[str, Any]):
    schema.pop("type", None)
    schema["anyOf"] = [{"type": "string"}, {"type": "boolean"}]


field_optional_bool_allowed_as_string_in_schema = Field(
    default=False, json_schema_extra=set_any_of_string_boolean
)


class Triggers(BaseModel):
    github_webhook: bool = field_optional_bool_allowed_as_string_in_schema
    git_provider_webhook: Optional[bool] = field_optional_bool_allowed_as_string_in_schema
    schedule: Optional[bool] = field_optional_bool_allowed_as_string_in_schema
    on_merge: Optional[bool] = field_optional_bool_allowed_as_string_in_schema


class Settings(BaseModel):
    threads: int = 4
    target_name: str = "default"


class Date(BaseModel):
    type: str
    cron: Optional[str] = None
    # TODO: Redo this one!
    # days: Optional[List[int]]


class Time(BaseModel):
    type: str
    interval: Optional[int] = None
    # TODO: Add this field back in
    # hours: Optional[List[int]] = []

    def serialize(self):
        payload: Dict[str, Any] = {"type": self.type}
        if self.type == "every_hour":
            payload["interval"] = self.interval
        elif self.type == "at_exact_hours":
            payload["hours"] = self.hours  # type: ignore
        return payload


class Schedule(BaseModel):
    cron: str
    date: Optional[Date] = None
    time: Optional[Time] = None

    def __init__(self, **data: Any):
        """Defaults to the same value as the UI."""
        data["date"] = Date(type="custom_cron", cron=data["cron"])
        data["time"] = Time(type="every_hour", interval=1)
        super().__init__(**data)

    @field_validator("cron")
    def valid_cron(cls, v):
        if not croniter.is_valid(v):
            raise ValueError("The cron expression is not valid")
        return v

    @field_serializer("time", when_used="json")
    def serialize_field(time: Optional[Time]):  # type: ignore
        if time is None:
            return None
        return time.serialize()


class Condition(BaseModel):
    job_id: int = field_mandatory_int_allowed_as_string_in_schema
    project_id: int = field_mandatory_int_allowed_as_string_in_schema
    statuses: List[Literal[10, 20, 30]] = Field(
        default=[10, 20, 30],
        description="The statuses that will trigger the job. 10=success 20=error 30=cancelled",
    )


class JobCompletionTriggerCondition(BaseModel):
    condition: Condition
